// Copyright (c) 2023 Contributors to the Eclipse Foundation
//
// See the NOTICE file(s) distributed with this work for additional
// information regarding copyright ownership.
//
// This program and the accompanying materials are made available under the
// terms of the Apache Software License 2.0 which is available at
// https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
// which is available at https://opensource.org/licenses/MIT.
//
// SPDX-License-Identifier: Apache-2.0 OR MIT

pub const POSIX_SUPPORT_ACL: bool = false;
pub const POSIX_SUPPORT_NAMED_SEMAPHORE: bool = false;
pub const POSIX_SUPPORT_PERSISTENT_SHARED_MEMORY: bool = false;
pub const POSIX_SUPPORT_UNIX_DATAGRAM_SOCKETS_ANCILLARY_DATA: bool = false;
pub const POSIX_SUPPORT_CPU_AFFINITY: bool = true;
pub const POSIX_SUPPORT_USERS_AND_GROUPS: bool = false;
pub const POSIX_SUPPORT_PERMISSIONS: bool = true;
pub const POSIX_SUPPORT_FILE_LOCK: bool = false;
pub const POSIX_SUPPORT_MEMORY_LOCK: bool = false;
pub const POSIX_SUPPORT_MESSAGE_QUEUE: bool = false;
pub const POSIX_SUPPORT_ADVANCED_SIGNAL_HANDLING: bool = false;
pub const POSIX_SUPPORT_CONSOLE_SIGNAL_HANDLING: bool = true;
pub const POSIX_SUPPORT_SCHEDULER: bool = false;
